from datasets import load_dataset
from torch.utils.data import DataLoader
from transformers import BertModel, BertTokenizerFast

from ckatorch import CKA


def tokenize(sentences_to_tokenize: dict[str, str]):
    return bert_tokenizer(
        sentences_to_tokenize["sentence1"],
        sentences_to_tokenize["sentence2"],
        truncation=True,
        add_special_tokens=True,
        padding="longest",
        max_length=bert_tokenizer.model_max_length,
        return_tensors="pt",
    ).to("cuda:0")


if __name__ == "__main__":
    # Load the dataset
    dataset = load_dataset("nyu-mll/glue", "mrpc", split="test", verification_mode="no_checks")

    # Load the tokenizer and the model from the hub
    bert_tokenizer = BertTokenizerFast.from_pretrained("google-bert/bert-base-uncased")
    bert_config = BertModel.from_pretrained("google-bert/bert-base-uncased").config
    first_model = BertModel(config=bert_config)
    second_model = BertModel(config=bert_config)

    # Define the dataloader
    dataset.set_transform(tokenize, ["sentence1", "sentence2"])
    dataloader = DataLoader(dataset, batch_size=8, num_workers=0, shuffle=True)

    # Observable_layers
    layers = [f"encoder.layer.{i}.output.dense" for i in range(bert_config.num_hidden_layers)]

    # Build the CKA objects, one for confronting a model with itself and the other for comparing two different models
    cka_same_model = CKA(
        first_model=first_model,
        second_model=first_model,
        layers=layers,
        first_name="Bert",
        device="cuda:0",
    )
    cka_different_models = CKA(
        first_model=first_model,
        second_model=second_model,
        layers=layers,
        first_name="Bert_0",
        second_name="Bert_1",
        device="cuda:0",
    )

    # Compute the CKA values for both scenarios
    cka_matrix_same = cka_same_model(dataloader, epochs=1)
    cka_matrix_different = cka_different_models(dataloader, epochs=1)

    # Plot the CKA values
    plot_parameters = {
        # "show_ticks_labels": True,
        # "short_tick_labels_splits": 4,
        "use_tight_layout": True,
        "show_half_heatmap": True,
    }
    cka_same_model.plot_cka(
        cka_matrix=cka_matrix_same,
        title=f"Model {cka_same_model.first_model_info.name} compared with itself",
        **plot_parameters,
    )
    cka_different_models.plot_cka(cka_matrix=cka_matrix_different, **plot_parameters)



"""Utilities for computing and centering Gram matrices."""

import torch


def linear_kernel(x: torch.Tensor) -> torch.Tensor:
    """Computes the Gram (kernel) matrix for a linear kernel.

    Adapted from the one made by Kornblith et al.
    https://github.com/google-research/google-research/tree/master/representation_similarity.

    Args:
        x (torch.Tensor): tensor of shape (n, m).

    Returns:
        torch.Tensor: a Gram matrix which is a tensor of shape (n, n).
    """
    return torch.mm(x, x.T)


def rbf_kernel(x: torch.Tensor, threshold: float = 1.0) -> torch.Tensor:
    """Computes the Gram (kernel) matrix for an RBF kernel.

    Adapted from the one made by Kornblith et al.
    https://github.com/google-research/google-research/tree/master/representation_similarity.

    Args:
        x (torch.Tensor): tensor of shape (n, m).
        threshold (float): fraction of median Euclidean distance to use as RBF kernel bandwidth (default=1.0).

    Returns:
        torch.Tensor: a Gram matrix which is a tensor of shape (n, n).
    """
    dot_products = torch.mm(x, x.T)
    sq_norms = torch.diag(dot_products)
    sq_distances = -2 * dot_products + sq_norms[:, None] + sq_norms[None, :]
    sq_median_distance = torch.median(sq_distances)
    return torch.exp(-sq_distances / (2 * threshold**2 * sq_median_distance))


def center_gram_matrix(gram_matrix: torch.Tensor, unbiased: bool = False) -> torch.Tensor:
    """Centers a given Gram matrix.

    Adapted from the one made by Kornblith et al.
    https://github.com/google-research/google-research/tree/master/representation_similarity.

    Args:
        gram_matrix (torch.Tensor): tensor of shape (n, n).
        unbiased (bool): whether to use the unbiased version of the centering (default=False).

    Returns:
        torch.Tensor: the centered version of the given Gram matrix.
    """
    if not torch.allclose(gram_matrix, gram_matrix.T):
        raise ValueError("The given matrix must be symmetric.")

    gram_matrix = gram_matrix.detach().clone()
    if unbiased:
        n = gram_matrix.shape[0]
        gram_matrix.fill_diagonal_(0)
        means = torch.sum(gram_matrix, dim=0, dtype=torch.float64) / (n - 2)
        means -= torch.sum(means) / (2 * (n - 1))
        gram_matrix -= means[:, None]
        gram_matrix -= means[None, :]
        gram_matrix.fill_diagonal_(0)
    else:
        means = torch.mean(gram_matrix, dim=0, dtype=torch.float64)
        means -= torch.mean(means) / 2
        gram_matrix -= means[:, None]
        gram_matrix -= means[None, :]

    return gram_matrix



"""Utilities for plotting the CKA matrices."""

import matplotlib.pyplot as plt
import seaborn as sn
import torch


def plot_cka(
    cka_matrix: torch.Tensor,
    first_layers: list[str],
    second_layers: list[str],
    first_name: str = "First Model",
    second_name: str = "Second Model",
    save_path: str | None = None,
    title: str | None = None,
    vmin: float | None = None,
    vmax: float | None = None,
    cmap: str = "magma",
    show_ticks_labels: bool = False,
    short_tick_labels_splits: int | None = None,
    use_tight_layout: bool = True,
    show_annotations: bool = True,
    show_img: bool = True,
    show_half_heatmap: bool = False,
    invert_y_axis: bool = True,
) -> None:
    """Plot the CKA matrix obtained by calling CKA class __call__() method.

    Args:
        cka_matrix (torch.Tensor): the CKA matrix.
        first_layers (list[str]): list of the names of the first model's layers.
        second_layers (list[str]): list of the names of the second model's layers.
        first_name (str): name of the first model (default="First Model").
        second_name (str): name of the second model (default="Second Model").
        save_path (str | None): where to save the plot, if None then the plot will not be saved (default=None).
        title (str | None): the plot title, if None then a simple text with the name of both models will be used
            (default=None).
        vmin (float | None): values to anchor the colormap, otherwise they are inferred from the data and other keyword
            arguments (default=None).
        vmax (float | None): values to anchor the colormap, otherwise they are inferred from the data and other keyword
            arguments (default=None).
        cmap (str): the name of the colormap to use (default="magma").
        show_ticks_labels (bool): whether to show the tick labels (default=False).
        short_tick_labels_splits (int | None): only works when show_tick_labels is True. If it is not None, the tick
            labels will be shortened to the defined sublayer starting from the deepest level. E.g.: if the layer name
            is 'encoder.ff.linear' and this parameter is set to 1, then only 'linear' will be printed on the heatmap
            (default=None).
        use_tight_layout (bool): whether to use a tight layout in order not to cut any label in the plot (default=True).
        show_annotations (bool): whether to show the annotations on the heatmap (default=True).
        show_img (bool): whether to show the plot (default=True).
        show_half_heatmap (bool): whether to mask the upper left part of the heatmap since those valued are duplicates
            (default=False).
        invert_y_axis (bool): whether to invert the y-axis of the plot (default=True).

    Raises:
        ValueError: if ``vmax`` or ``vmin`` are not defined together or both equal to None.
    """
    # Deal with vmin and vmax
    if (vmin is not None) ^ (vmax is not None):
        raise ValueError("'vmin' and 'vmax' must be defined together or both equal to None.")

    vmin = min(vmin, torch.min(cka_matrix).item()) if vmin is not None else vmin
    vmax = max(vmax, torch.max(cka_matrix).item()) if vmax is not None else vmax

    # Build the mask
    if show_half_heatmap:
        mask = torch.tril(torch.ones_like(cka_matrix, dtype=torch.bool), diagonal=-1).cpu().numpy()
    else:
        mask = None

    # Build the heatmap
    ax = sn.heatmap(cka_matrix.cpu(), vmin=vmin, vmax=vmax, annot=show_annotations, cmap=cmap, mask=mask)
    if invert_y_axis:
        ax.invert_yaxis()

    ax.set_xlabel(f"{second_name} layers", fontsize=12)
    ax.set_ylabel(f"{first_name} layers", fontsize=12)

    # Deal with tick labels
    if show_ticks_labels:
        if short_tick_labels_splits is None:
            ax.set_xticklabels(second_name)
            ax.set_yticklabels(first_name)
        else:
            ax.set_xticklabels(["-".join(module.split(".")[-short_tick_labels_splits:]) for module in second_layers])
            ax.set_yticklabels(["-".join(module.split(".")[-short_tick_labels_splits:]) for module in first_layers])

        plt.xticks(rotation=90)
        plt.yticks(rotation=0)

    # Put the title if passed
    if title is not None:
        ax.set_title(title, fontsize=14)
    else:
        title = f"{first_name} vs {second_name}"
        ax.set_title(title, fontsize=14)

    # Set the layout to tight if the corresponding parameter is True
    if use_tight_layout:
        plt.tight_layout()

    # Save the plot to the specified path if defined
    if save_path is not None:
        title = title.replace(" ", "_").replace("/", "-")
        path_rel = f"{save_path}/{title}.png"
        plt.savefig(path_rel, dpi=400, bbox_inches="tight")

    # Show the image if the user chooses to do so
    if show_img:
        plt.show()



"""Module for computing CKA in order to compare two PyTorch models through their layers' activations."""

import inspect
import json
import random
import re
import string
from collections.abc import Callable
from dataclasses import asdict, dataclass
from functools import partial
from pathlib import Path
from typing import Any, Literal
from warnings import warn

import torch
import yaml
from safetensors.torch import save_file, save_model
from torch import nn
from torch.utils.data import DataLoader, RandomSampler
from tqdm import tqdm

from .core import cka_batch
from .plot import plot_cka


@dataclass
class ModelInfo:
    """Class for storing the model info."""

    name: str
    layers: list[str]


class CKA:
    """Centered Kernel Alignment (CKA) implementation.

    CKA is a similarity index between representations of features in neural networks, based on the
    Hilbert-Schmidt Independence Criterion (HSIC). Given a set of examples, CKA compares the representations of examples
    passed through the layers that we want to compare.
    """

    def __init__(
        self,
        first_model: nn.Module,
        second_model: nn.Module,
        layers: list[str],
        second_layers: list[str] | None = None,
        first_name: str | None = None,
        second_name: str | None = None,
        device: str | torch.device = "cpu",
    ) -> None:
        """Initializes a CKA object.

        Args:
            first_model (torch.nn.Module): the first model whose layer features we want to compare.
            second_model (torch.nn.Module): the second model whose layer features we want to compare.
            layers (list[str]): list of layers name under inspection (if no "second_layers" is provided, then the
                layers will count for the second model too).
            second_layers (list[str] | None): list of layers from the second model under inspection (default=None).
            first_name (str | None): name of the first model (default=None).
            second_name (str | None): name of the second model (default=None).
            device (str | torch.device): the device used during the computation (default="cpu").

        Raises:
            ValueError: if ``layers`` is None or empty.
        """
        # Set up the device
        self.device = torch.device(device)

        # Check if no layers were passed
        if layers is None or len(layers) == 0:
            raise ValueError(
                "You can not pass 'None' or an empty list as layers. We suggest using 'first_model.named_modules()'"
                "in order to see which layers can be passed."
            )

        # Remove potential duplicates
        layers = sorted(set(layers), key=layers.index)

        # Check if to many layers were passed
        if len(layers) > 100:
            warn(
                f"You passed {len(layers)} distinct layers, which is way too high. Consider passing only those"
                f"layers whose features you are really interested about."
            )

        # Copy the first model's layers if they are not passed
        if second_layers is None or len(second_layers) == 0:
            second_layers = layers.copy()
        else:
            # Remove potential duplicates
            second_layers = sorted(set(second_layers), key=second_layers.index)

            # Check if too many layers were passed
            if len(second_layers) > 100:
                warn(
                    f"You passed {len(second_layers)} distinct layers for the second model, which is way too high."
                    f"Consider passing only those layers whose features you are really interested about."
                )

        # Dicts where the output of each layer (i.e.: the features) will be saved while using hooks
        self.first_features: dict[str, torch.Tensor] = {}
        self.second_features: dict[str, torch.Tensor] = {}

        # Insert a hook for each layer
        layers, second_layers = self._insert_hooks(first_model, second_model, layers, second_layers)
        self.first_model = first_model.to(device)
        self.second_model = second_model.to(device)

        # Manage the models names
        first_name = first_name if first_name is not None else first_model.__repr__().split("(")[0]
        if first_model is second_model:
            second_name = first_name
        else:
            second_name = second_name if second_name is not None else second_model.__repr__().split("(")[0]
            if first_name == second_name:
                warn(f"Both models are called {first_name}. This may cause confusion when analyzing the results.")

        # Set up the models info
        first_name = re.sub("[^0-9a-zA-Z_]+", "", first_name.replace(" ", "_"))
        second_name = re.sub("[^0-9a-zA-Z_]+", "", second_name.replace(" ", "_"))
        self.first_model_info = ModelInfo(first_name, layers)
        self.second_model_info = ModelInfo(second_name, second_layers)

    def _hook(self, model: str, module_name: str, module: nn.Module, inp: torch.Tensor, out: torch.Tensor) -> None:
        del module, inp  # delete unused arguments so that we can pass the linter checks
        if model == "first":
            self.first_features[module_name] = out.detach()
        else:
            self.second_features[module_name] = out.detach()

    def _insert_hooks(
        self,
        first_model: nn.Module,
        second_model: nn.Module,
        layers: list[str],
        second_layers: list[str],
    ) -> tuple[list[str], list[str]]:
        # Only those layers that were found will be placed inside the following lists
        filtered_first_layers = []
        filtered_second_layers = []

        # Add hooks for the first model's layers
        for module_name, module in list(first_model.named_modules()):
            if module_name in layers:
                module.register_forward_hook(partial(self._hook, "first", module_name))
                filtered_first_layers.append(module_name)

        # Add hooks for the second model's layers
        for module_name, module in list(second_model.named_modules()):
            if module_name in second_layers:
                module.register_forward_hook(partial(self._hook, "second", module_name))
                filtered_second_layers.append(module_name)

        # One last check
        if len(filtered_first_layers) == 0 or len(filtered_second_layers) == 0:
            raise ValueError(
                "No layers were found for one of the two models, please use 'model.named_modules()' in order to check"
                "which layers can be passed to the method."
            )

        return filtered_first_layers, filtered_second_layers

    def __call__(
        self,
        dataloader: DataLoader,
        epochs: int = 10,
        f_extract: Callable[..., dict[str, torch.Tensor]] | None = None,
        f_args: dict[str, Any] | None = None,
    ) -> torch.Tensor:
        """Process inputs and computes the CKA matrix.

        This computation employs the minibatch version of CKA by Nguyen et al. (https://arxiv.org/abs/2010.15327).

        Args:
            dataloader (torch.utils.data.Dataloader): dataloader that will be used during the computation.
            epochs (int): number of iterations over the dataloader (default=10).
            f_extract (Callable[..., dict[str, torch.Tensor]] | None): the function to apply on the dataloader, this
                function should take any number and type of inputs and return a dict. If no function is passed, then
                some checks will be applied for finding the actual type of the batch (default=None).
            f_args (dict[str, Any] | None): the arguments passed to the f_extract function (default=None).

        Returns:
            torch.Tensor: a tensor representing the CKA matrix.

        Raises:
            ValueError: if the parameter 'drop_last' of the dataloader is set to True or if the batch type is not
                supported.
        """
        if dataloader.drop_last:
            raise ValueError(
                "The argument 'drop_last' must be set to False otherwise you will get very different values by varying"
                "the batch size."
            )

        if not isinstance(dataloader.sampler, RandomSampler):
            warn("We suggest setting 'shuffle=True' in your dataloader in order to have a less biased computation.")

        self.first_model.eval()
        self.second_model.eval()

        with torch.no_grad():
            n = len(self.first_model_info.layers)
            m = len(self.second_model_info.layers)

            # Iterate through the dataset
            num_batches = len(dataloader)
            cka_matrices = []
            for epoch in tqdm(range(epochs), desc="| Computing CKA |", total=epochs):
                cka_epoch = torch.zeros(n, m, device=self.device)
                for batch in tqdm(dataloader, desc=f"| Computing CKA epoch {epoch} |", total=num_batches, leave=False):
                    self.first_features = {}
                    self.second_features = {}
                    if f_extract is not None:
                        # Apply the provided function and put everything on the device
                        f_extract = {} if f_extract is None else f_extract
                        batch = f_extract(batch, **f_args)
                        batch = {f"{name}": batch_input.to(self.device) for name, batch_input in batch.items()}
                    elif isinstance(batch, list | tuple):
                        arg_method = self.first_model.forward
                        args_list = inspect.getfullargspec(arg_method).args[1:]  # skip "self" arg
                        batch = {f"{args_list[i]}": batch_input.to(self.device) for i, batch_input in enumerate(batch)}
                    elif not isinstance(batch, dict):
                        raise ValueError(
                            f"Type {type(batch)} is not supported for the CKA computation. We suggest building a custom"
                            f"'Dataset' class such that the '__get_item__' method returns a dict[str, Any]."
                        )

                    # Do a forward pass for both models
                    _ = self.first_model(**batch)
                    _ = self.second_model(**batch)
                    first_outputs = self.first_features
                    second_outputs = self.second_features

                    # Compute the CKA values for each output pair
                    for i, (_, x) in enumerate(first_outputs.items()):
                        for j, (_, y) in enumerate(second_outputs.items()):
                            cka_epoch[i, j] = cka_batch(x, y)

                cka_matrices.append(cka_epoch)

        cka = torch.stack(cka_matrices)
        cka = cka.sum(0) / epochs

        # One last check
        if torch.isnan(cka).any():
            raise ValueError("CKA computation resulted in NANs.")

        return cka

    def plot_cka(
        self,
        cka_matrix: torch.Tensor,
        save_path: str | None = None,
        title: str | None = None,
        vmin: float | None = None,
        vmax: float | None = None,
        cmap: str = "magma",
        show_ticks_labels: bool = False,
        short_tick_labels_splits: int | None = None,
        use_tight_layout: bool = True,
        show_annotations: bool = True,
        show_img: bool = True,
        show_half_heatmap: bool = False,
        invert_y_axis: bool = True,
    ) -> None:
        """Plot the CKA matrix obtained by calling this class' __call__() method.

        Args:
            cka_matrix (torch.Tensor): the CKA matrix.
            save_path (str | None): where to save the plot, if None then the plot will not be saved (default=None).
            title (str | None): the plot title, if None then a simple text with the name of both models will be used
                (default=None).
            vmin (float | None): values to anchor the colormap, otherwise they are inferred from the data and other
                keyword arguments (default=None).
            vmax (float | None): values to anchor the colormap, otherwise they are inferred from the data and other
                keyword arguments (default=None).
            cmap (str): the name of the colormap to use (default="magma").
            show_ticks_labels (bool): whether to show the tick labels (default=False).
            short_tick_labels_splits (int | None): only works when show_tick_labels is True. If it is not None, the
                tick labels will be shortened to the defined sublayer starting from the deepest level. E.g.: if the
                layer name is 'encoder.ff.linear' and this parameter is set to 1, then only 'linear' will be printed on
                the heatmap (default=None).
            use_tight_layout (bool): whether to use a tight layout in order not to cut any label in the plot
                (default=True).
            show_annotations (bool): whether to show the annotations on the heatmap (default=True).
            show_img (bool): whether to show the plot (default=True).
            show_half_heatmap (bool): whether to mask the upper left part of the heatmap since those valued are
                duplicates (default=False).
            invert_y_axis (bool): whether to invert the y-axis of the plot (default=True).


        Raises:
            ValueError: if ``vmax`` or ``vmin`` are not defined together or both equal to None.
        """
        plot_cka(
            cka_matrix=cka_matrix,
            first_layers=self.first_model_info.layers,
            second_layers=self.second_model_info.layers,
            first_name=self.first_model_info.name,
            second_name=self.second_model_info.name,
            save_path=save_path,
            title=title,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            show_ticks_labels=show_ticks_labels,
            short_tick_labels_splits=short_tick_labels_splits,
            use_tight_layout=use_tight_layout,
            show_annotations=show_annotations,
            show_img=show_img,
            show_half_heatmap=show_half_heatmap,
            invert_y_axis=invert_y_axis,
        )

    def save(
        self,
        cka_matrix: torch.Tensor,
        dir_path: str | Path = "",
        file_format: Literal["json", "yaml"] = "json",
        save_models: bool = False,
    ) -> None:
        """Saves the CKA matrix and the info about the models used during the computation.

        Note that the CKA matrix and the models' weights are saved as safetensors files and their paths are be stored
        inside the json or yaml file.

        Args:
            cka_matrix (torch.Tensor): the CKA matrix.
            dir_path (str | pathlib.Path): where to save the weights of the models and the file with their info
                (default="").
            file_format (Literal["json", "yaml"]): in which format the output is saved, can be either 'json' or 'yaml'
                (default: "json").
            save_models (bool): whether to also save the models used for the CKA computation (default=False).

        Raises:
            ValueError: if ``file_format`` not in ['json', 'yaml'].
        """
        # Save the CKA matrix
        dir_path = Path(dir_path)
        dir_path.mkdir(parents=True, exist_ok=True)
        cka_tensor_path = dir_path / "cka.safetensors"
        cka_info_path = dir_path / f"cka.{file_format}"
        random_str = ""
        if cka_tensor_path.exists():
            # If the file already exists, change the name by appending five random characters
            random_str = "_" + "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(5))
            cka_tensor_path = dir_path / f"cka{random_str}.safetensors"
            cka_info_path = dir_path / f"cka{random_str}.{file_format}"

        save_file({"cka": cka_matrix}, cka_tensor_path)

        # Build the dict to dump
        cka_info = {
            "cka_matrix": {
                "path": str(cka_tensor_path.absolute()),
            },
            "first_model": asdict(self.first_model_info),
            "second_model": asdict(self.second_model_info),
        }

        # Save the models if requested
        if save_models:
            first_model_path = dir_path / f"{self.first_model_info.name}{random_str}.safetensors"
            second_model_path = dir_path / f"{self.second_model_info.name}{random_str}.safetensors"
            save_model(self.first_model, first_model_path)
            save_model(self.second_model, second_model_path)
            cka_info["first_model"]["path"] = str(first_model_path.absolute())
            cka_info["second_model"]["path"] = str(second_model_path.absolute())

        # Dump the dict
        with open(cka_info_path, "w", encoding="utf-8") as dump_file:
            match file_format:
                case "json":
                    json.dump(cka_info, dump_file, indent=4)
                case "yaml":
                    yaml.dump(cka_info, dump_file, indent=4)
                case _:
                    raise ValueError(f"Unsupported format '{file_format}'")



"""Module for computing HSIC (Hilbert-Schmidt Independence Criterion), both its standard and mini-batch versions."""

import torch


def hsic0(gram_x: torch.Tensor, gram_y: torch.Tensor) -> torch.Tensor:
    """Compute the Hilbert-Schmidt Independence Criterion on two given Gram matrices.

    Args:
        gram_x (torch.Tensor): Gram matrix of shape (n, n), this is equivalent to K from the original paper.
        gram_y (torch.Tensor): Gram matrix of shape (n, n), this is equivalent to L from the original paper.

    Returns:
        torch.Tensor: a tensor with the Hilbert-Schmidt Independence Criterion values.

    Raises:
        ValueError: if ``gram_x`` and ``gram_y`` are not symmetric.
    """
    if not torch.allclose(gram_x, gram_x.T) and not torch.allclose(gram_y, gram_y.T):
        raise ValueError("The given matrices must be symmetric.")

    # Build the identity matrix
    n = gram_x.shape[0]
    identity = torch.eye(n, n, dtype=gram_x.dtype, device=gram_x.device)

    # Build the centering matrix
    h = identity - torch.ones(n, n, dtype=gram_x.dtype, device=gram_x.device) / n

    # Compute k * h and l * h
    kh = torch.mm(gram_x, h)
    lh = torch.mm(gram_y, h)

    # Compute the trace of the product kh * lh
    trace = torch.trace(kh.mm(lh))
    return trace / (n - 1) ** 2


def hsic1(gram_x: torch.Tensor, gram_y: torch.Tensor) -> torch.Tensor:
    """Compute the batched version of the Hilbert-Schmidt Independence Criterion on Gram matrices.

    This version is based on
    https://github.com/numpee/CKA.pytorch/blob/07874ec7e219ad29a29ee8d5ebdada0e1156cf9f/cka.py#L107.

    Args:
        gram_x (torch.Tensor): batch of Gram matrices of shape (bsz, n, n).
        gram_y (torch.Tensor): batch of Gram matrices of shape (bsz, n, n).

    Returns:
        torch.Tensor: a tensor with the unbiased Hilbert-Schmidt Independence Criterion values.

    Raises:
        ValueError: if ``gram_x`` and ``gram_y`` do not have the same shape or if they do not have exactly three
            dimensions.
    """
    if len(gram_x.size()) != 3 or gram_x.size() != gram_y.size():
        raise ValueError("Invalid size for one of the two input tensors.")

    n = gram_x.shape[-1]
    gram_x = gram_x.clone()
    gram_y = gram_y.clone()

    # Fill the diagonal of each matrix with 0
    gram_x.diagonal(dim1=-1, dim2=-2).fill_(0)
    gram_y.diagonal(dim1=-1, dim2=-2).fill_(0)

    # Compute the product between k (i.e.: gram_x) and l (i.e.: gram_y)
    kl = torch.bmm(gram_x, gram_y)

    # Compute the trace (sum of the elements on the diagonal) of the previous product, i.e.: the left term
    trace_kl = kl.diagonal(dim1=-1, dim2=-2).sum(-1).unsqueeze(-1).unsqueeze(-1)

    # Compute the middle term
    middle_term = gram_x.sum((-1, -2), keepdim=True) * gram_y.sum((-1, -2), keepdim=True)
    middle_term /= (n - 1) * (n - 2)

    # Compute the right term
    right_term = kl.sum((-1, -2), keepdim=True)
    right_term *= 2 / (n - 2)

    # Put all together to compute the main term
    main_term = trace_kl + middle_term - right_term

    # Compute the hsic values
    out = main_term / (n**2 - 3 * n)
    return out.squeeze(-1).squeeze(-1)



"""Module that implements both base and mini-batch CKA."""

from typing import Literal

import torch

from .hsic import hsic0, hsic1
from .utils import center_gram_matrix, linear_kernel, rbf_kernel


def cka_base(
    x: torch.Tensor,
    y: torch.Tensor,
    kernel: Literal["linear", "rbf"] = "linear",
    unbiased: bool = False,
    threshold: float = 1.0,
    method: Literal["fro_norm", "hsic"] = "fro_norm",
) -> torch.Tensor:
    """Computes the Centered Kernel Alignment (CKA) between two given matrices.

    Adapted from the one made by Kornblith et al.
    https://github.com/google-research/google-research/tree/master/representation_similarity.

    Args:
        x (torch.Tensor): tensor of shape (n, j).
        y (torch.Tensor): tensor of shape (n, k).
        kernel (Literal["linear", "rbf"]): the kernel used to compute the Gram matrices, must be "linear" or "rbf"
            (default="linear").
        unbiased (bool): whether to use the unbiased version of CKA (default=False).
        threshold (float): the threshold used by the RBF kernel (default=1.0).
        method (Literal["fro_norm", "hsic"]): the method used to compute the CKA value, must be "fro_norm"
            (Frobenius norm) or "hsic" (Hilbert-Schmidt Independence Criterion). Note that the choice does not
            influence the output (default="fro_norm").

    Returns:
        torch.Tensor: a float tensor in [0, 1] that is the CKA value between the two given matrices.

    Raises:
        ValueError: if ``kernel`` is not "linear" or "rbf" or if ``method`` is not "fro_norm" or "hsic".
    """
    if kernel not in ["linear", "rbf"]:
        raise ValueError("The chosen kernel must be either 'linear' or 'rbf'.")

    if method not in ["hsic", "fro_norm"]:
        raise ValueError("The chosen method must be either 'hsic' or 'fro_norm'.")

    x = x.type(torch.float64) if not x.dtype == torch.float64 else x
    y = y.type(torch.float64) if not y.dtype == torch.float64 else y

    # Build the Gram matrices by applying the kernel
    gram_x = linear_kernel(x) if kernel == "linear" else rbf_kernel(x, threshold)
    gram_y = linear_kernel(y) if kernel == "linear" else rbf_kernel(y, threshold)

    # Compute CKA by either using HSIC or the Frobenius norm
    if method == "hsic":
        hsic_xy = hsic0(gram_x, gram_y)
        hsic_xx = hsic0(gram_x, gram_x)
        hsic_yy = hsic0(gram_y, gram_y)
        cka = hsic_xy / torch.sqrt(hsic_xx * hsic_yy)
    else:
        gram_x = center_gram_matrix(gram_x, unbiased)
        gram_y = center_gram_matrix(gram_y, unbiased)
        norm_xy = gram_x.ravel().dot(gram_y.ravel())
        norm_xx = torch.linalg.norm(gram_x, ord="fro")
        norm_yy = torch.linalg.norm(gram_y, ord="fro")
        cka = norm_xy / (norm_xx * norm_yy)

    return cka


def cka_batch(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Compute the minibatch version of CKA from Nguyen et al. (https://arxiv.org/abs/2010.15327).

    This computation is performed with linear kernel and by calculating HSIC_1.

    Args:
        x (torch.Tensor): tensor of shape (bsz, n, j).
        y (torch.Tensor): tensor of shape (bsz, n, k).

    Returns:
        torch.Tensor: a float tensor in [0, 1] that is the CKA value between the two given tensors.
    """
    x = x.type(torch.float64) if not x.dtype == torch.float64 else x
    y = y.type(torch.float64) if not y.dtype == torch.float64 else y

    # Build the Gram matrices by applying the linear kernel
    gram_x = torch.bmm(x, x.transpose(1, 2))
    gram_y = torch.bmm(y, y.transpose(1, 2))

    # Compute the HSIC values for the entire batches
    hsic1_xy = hsic1(gram_x, gram_y)
    hsic1_xx = hsic1(gram_x, gram_x)
    hsic1_yy = hsic1(gram_y, gram_y)

    # Compute the CKA value
    cka = hsic1_xy.sum() / (hsic1_xx.sum() * hsic1_yy.sum()).sqrt()
    return cka



"""Centered Kernel Alignment in PyTorch."""

from .cka import CKA
from .core import cka_base, cka_batch
from .hsic import hsic0, hsic1
from .plot import plot_cka
from .utils import center_gram_matrix, linear_kernel, rbf_kernel


__all__ = [
    "CKA",
    "center_gram_matrix",
    "cka_base",
    "cka_batch",
    "hsic0",
    "hsic1",
    "linear_kernel",
    "plot_cka",
    "rbf_kernel",
]



import json

from safetensors.torch import load_file

from ckatorch.plot import plot_cka


if __name__ == "__main__":
    # Load the info file
    with open("cka.json") as info_file:
        model_info = json.load(info_file)

    # Load the CKA matrix
    cka_matrix = load_file(model_info["cka_matrix"]["path"])["cka"]

    # Plot the CKA values
    plot_cka(
        cka_matrix=cka_matrix,
        first_layers=model_info["first_model"]["layers"],
        second_layers=model_info["second_model"]["layers"],
        save_path=".",
        show_ticks_labels=True,
        short_tick_labels_splits=2,
        use_tight_layout=True,
        show_half_heatmap=True,
    )



import torch
from torch import nn
from torch.functional import F
from torch.utils.data import DataLoader, TensorDataset

from ckatorch import CKA


class FF(nn.Module):
    def __init__(self) -> None:
        """Simple feed-forward network like the ones used in the transformer architecture."""
        super().__init__()
        self.linear1 = nn.Linear(512, 2048, dtype=torch.float64)
        self.dropout = nn.Dropout(0.0)
        self.linear2 = nn.Linear(2048, 512, dtype=torch.float64)
        self.activation = F.relu

    def forward(self, src: torch.Tensor) -> torch.Tensor:
        out = self.activation(self.linear1(src))
        out = self.dropout(out)
        out = self.linear2(out)
        return out


class ManyFF(nn.Module):
    def __init__(self) -> None:
        """Many feed-forward networks one after the other."""
        super().__init__()
        self.model = nn.Sequential(FF(), FF(), FF(), FF(), FF(), FF())

    def forward(self, src: torch.Tensor) -> torch.Tensor:
        return self.model(src)


if __name__ == "__main__":
    # Set a seed for reproducibility
    torch.manual_seed(5)

    # Build the models
    first_many_ff = ManyFF()
    second_many_ff = ManyFF()

    # Define the layers that will be used during the computation
    layers_to_observe = [
        "model.0.linear1",
        "model.1.linear1",
        "model.2.linear1",
        "model.3.linear1",
        "model.4.linear1",
        "model.5.linear1",
    ]

    # Define the shared parameters between the two CKA objects
    shared_parameters = {
        "layers": layers_to_observe,
        "first_name": "ManyFF_0",
        "device": torch.device("cuda:0") if torch.cuda.is_available() else "cpu",
    }

    # Build the CKA objects, one for confronting a model with itself and the other for comparing two different models
    cka_same_model = CKA(
        first_model=first_many_ff,
        second_model=first_many_ff,
        **shared_parameters,
    )
    cka_different_models = CKA(
        first_model=first_many_ff,
        second_model=second_many_ff,
        second_name="ManyFF_1",
        **shared_parameters,
    )

    # Build the dataset from a random three-dimensional tensor and define the dataloader
    dataset = TensorDataset(torch.randn(128, 256, 512, dtype=torch.float64))
    dataloader = DataLoader(dataset, batch_size=8, num_workers=8, shuffle=True)

    # Compute the CKA values for both scenarios
    cka_matrix_same = cka_same_model(dataloader)
    cka_matrix_different = cka_different_models(dataloader)

    # Plot the CKA values
    plot_parameters = {
        "show_ticks_labels": True,
        "short_tick_labels_splits": 2,
        "use_tight_layout": True,
        "show_half_heatmap": True,
    }
    cka_same_model.plot_cka(
        cka_matrix=cka_matrix_same,
        title=f"Model {cka_same_model.first_model_info.name} compared with itself",
        **plot_parameters,
    )
    cka_different_models.plot_cka(cka_matrix=cka_matrix_different, **plot_parameters)
